/*
 * RFBRectangle.m
 *
 *
 * Copyright (c) 1997 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1997.
 */

#import "VNC.h"

@implementation RFBRectangle

+ (int *)supportedEncodings
{
    static int supportedEncodings[] = {
//        RFBEncodingHextile,
        RFBEncodingCopyRect,
        RFBEncodingCoRRE,
        RFBEncodingRRE,
        0
    };

    return supportedEncodings;
}

+ (unsigned)numberOfSupportedEncodings
{
    return 3;
}

+ (RFBRectangle *)rectangleWithAdaptor:(RFBProtocolAdaptor *)adaptor
{
    TCStream *stream = [adaptor socket];
    NSRect r;
    unsigned encoding = 0;

    r.origin.x = [stream readUnsignedShort];
    r.origin.y = [stream readUnsignedShort];
    r.size.width = [stream readUnsignedShort];
    r.size.height = [stream readUnsignedShort];

    encoding = [stream readUnsignedInt];
    
    switch(encoding) {
        case RFBEncodingRaw:
            return [RFBRawRectangle rectangleWithLocation:r encoding:encoding adaptor:adaptor];

        case RFBEncodingCopyRect:
            return [RFBCopyRectangle rectangleWithLocation:r encoding:encoding adaptor:adaptor];

        case RFBEncodingRRE:
            return [RFBRRERectangle rectangleWithLocation:r encoding:encoding adaptor:adaptor];

        case RFBEncodingCoRRE:
            return [RFBCoRRERectangle rectangleWithLocation:r encoding:encoding adaptor:adaptor];

        case RFBEncodingHextile:
            return [RFBHextileRectangle rectangleWithLocation:r encoding:encoding adaptor:adaptor];

        default:
            [NSException raise:NSInvalidArgumentException
                        format:@"%@: unknown encoding method %d",
                NSStringFromSelector(_cmd), encoding];
    }

    return nil; // not reached
}

+ (RFBRectangle *)rectangleWithLocation:(NSRect)location encoding:(unsigned)encoding adaptor:(RFBProtocolAdaptor *)adaptor
{
    return [[[self alloc] initWithLocation:location encoding:encoding adaptor:adaptor] autorelease];
}

- (id)initWithLocation:(NSRect)location encoding:(unsigned)encoding adaptor:(RFBProtocolAdaptor *)adaptor
{
    if (self = [super init]) {
        _location.origin.x = location.origin.x;
        _location.origin.y = location.origin.y;
        _location.size.width = location.size.width;
        _location.size.height = location.size.height;

        _encoding = encoding;
    }

    return self;
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] location=%@ encoding=%d>",
        [[self class] description], self,
        NSStringFromRect(_location), _encoding];
}

- (void)drawWithFramebuffer:(VNCFramebuffer *)framebuffer
{
}

- (id)copyWithZone:(NSZone *)zone
{
    return [self retain];
}

- (NSRect)location
{
    return _location;
}

- (unsigned)encoding
{
    return _encoding;
}

- (id)initWithCoder:(NSCoder *)coder
{
    if (self = [super init]) {
        _location = [coder decodeRect];
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_encoding];
    }

    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [coder encodeRect:_location];
    [coder encodeValueOfObjCType:@encode(unsigned) at:&_encoding];
}

@end

@implementation RFBRawRectangle

- (id)initWithLocation:(NSRect)location encoding:(unsigned)encoding adaptor:(RFBProtocolAdaptor *)adaptor
{
    if (self = [super initWithLocation:location encoding:encoding adaptor:adaptor]) {
        unsigned imageLength = location.size.width * location.size.height * ([adaptor bitsPerPixel]/8);
        _bitsPerPixel = [adaptor bitsPerPixel];
        _pixelData = [[[adaptor socket] readDataOfLength:imageLength] retain];
    }

    return self;
}

- (void)drawWithFramebuffer:(VNCFramebuffer *)framebuffer
{
    [framebuffer compositePixelData:_pixelData
                         targetRect:_location
                       bitsPerPixel:_bitsPerPixel];
}

- (void)dealloc
{
    [_pixelData release];

    [super dealloc];
}

- (NSData *)pixelData
{
    return _pixelData;
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] location=%@ encoding=%d pixelData length=%u>",
        [[self class] description], self,
        NSStringFromRect(_location), _encoding, [_pixelData length]];
}

- (id)initWithCoder:(NSCoder *)coder
{
    if (self = [super initWithCoder:coder]) {
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_bitsPerPixel];
        _pixelData = [[coder decodeDataObject] retain];
    }

    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [super encodeWithCoder:coder];

    [coder encodeValueOfObjCType:@encode(unsigned) at:&_bitsPerPixel];
    [coder encodeDataObject:_pixelData];
}

@end

@implementation RFBCopyRectangle

- (id)initWithLocation:(NSRect)location encoding:(unsigned)encoding adaptor:(RFBProtocolAdaptor *)adaptor
{
    if (self = [super initWithLocation:location encoding:encoding adaptor:adaptor]) {
        _source = NSMakePoint([[adaptor socket] readUnsignedShort],
                              [[adaptor socket] readUnsignedShort]);
    }

    return self;
}

- (NSPoint)source
{
    return _source;
}

- (void)drawWithFramebuffer:(VNCFramebuffer *)framebuffer
{
    [framebuffer copyPixelsFromRect:NSMakeRect(_source.x, _source.y,
                                               _location.size.width, _location.size.height)
                             toRect:_location];
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] location=%@ encoding=%d source=%@>",
        [[self class] description], self,
        NSStringFromRect(_location), _encoding, NSStringFromPoint(_source)];
}

- (id)initWithCoder:(NSCoder *)coder
{
    if (self = [super initWithCoder:coder]) {
        _source = [coder decodePoint];
    }

    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [super encodeWithCoder:coder];
    [coder encodePoint:_source];
}

@end

@interface RFBRRERectangle(Private)
- (void)_readRectanglesWithAdaptor:(RFBProtocolAdaptor *)adaptor;
@end

@implementation RFBRRERectangle

- (id)initWithLocation:(NSRect)location encoding:(unsigned)encoding adaptor:(RFBProtocolAdaptor *)adaptor
{
    if (self = [super initWithLocation:location encoding:encoding adaptor:adaptor]) {
        _numberOfRects = [[adaptor socket] readUnsignedInt];
        _backgroundColor = [adaptor read24BitColor];

        _rectListCover = [[NSMutableData alloc] initWithLength:(_numberOfRects*sizeof(NSRect))];
        _colorListCover = [[NSMutableData alloc] initWithLength:(_numberOfRects*sizeof(unsigned))];

        [self _readRectanglesWithAdaptor:adaptor];
    }

    return self;
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] location=%@ encoding=%d numberOfRects=%d backgroundColor=0x%08lx>",
        [[self class] description], self,
        NSStringFromRect(_location), _encoding, _numberOfRects, _backgroundColor];
}

- (void)_readRectanglesWithAdaptor:(RFBProtocolAdaptor *)adaptor
{
    NSRect *rectList = [_rectListCover mutableBytes];
    unsigned *colorList = [_colorListCover mutableBytes];
    unsigned i;

    for (i = 0; i < _numberOfRects; ++i) {
        colorList[i] = [adaptor read24BitColor];
        rectList[i].origin.x = _location.origin.x + [[adaptor socket] readUnsignedShort];
        rectList[i].origin.y = _location.origin.y + [[adaptor socket] readUnsignedShort];
        rectList[i].size.width = [[adaptor socket] readUnsignedShort];
        rectList[i].size.height = [[adaptor socket] readUnsignedShort];
    }
}

- (void)dealloc
{
    [_rectListCover release];
    [_colorListCover release];

    [super dealloc];
}

- (void)drawWithFramebuffer:(VNCFramebuffer *)framebuffer
{
    [framebuffer fillRect:_location color:_backgroundColor];    
    [framebuffer fillRectList:[_rectListCover bytes]
                    colorList:[_colorListCover bytes]
                        count:_numberOfRects];
}

- (id)initWithCoder:(NSCoder *)coder
{
    if (self = [super initWithCoder:coder]) {
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_backgroundColor];
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_numberOfRects];

        _rectListCover = [[coder decodeDataObject] retain];
        _colorListCover = [[coder decodeDataObject] retain];        
    }

    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [super encodeWithCoder:coder];

    [coder encodeValueOfObjCType:@encode(unsigned) at:&_backgroundColor];
    [coder encodeValueOfObjCType:@encode(unsigned) at:&_numberOfRects];
    [coder encodeDataObject:_rectListCover];
    [coder encodeDataObject:_colorListCover];
}

@end

@implementation RFBCoRRERectangle

- (void)_readRectanglesWithAdaptor:(RFBProtocolAdaptor *)adaptor
{
    NSRect *rectList = [_rectListCover mutableBytes];
    unsigned *colorList = [_colorListCover mutableBytes];
    unsigned i;

    for (i = 0; i < _numberOfRects; ++i) {
        colorList[i] = [adaptor read24BitColor];
        rectList[i].origin.x = _location.origin.x + [[adaptor socket] readByte];
        rectList[i].origin.y = _location.origin.y + [[adaptor socket] readByte];
        rectList[i].size.width = [[adaptor socket] readByte];
        rectList[i].size.height = [[adaptor socket] readByte];
    }
}

@end

@implementation RFBHextileRectangle

- (id)initWithLocation:(NSRect)location encoding:(unsigned)encoding adaptor:(RFBProtocolAdaptor *)adaptor
{
    if (self = [super initWithLocation:location encoding:encoding adaptor:adaptor]) {
        NSRect subRect;
        NSPoint maxPoint = NSMakePoint(_location.origin.x+_location.size.width,
                                       _location.origin.y+_location.size.height);
        
        _subrectangles = [[NSMutableArray alloc] init];

        for (subRect.origin.y = _location.origin.y; subRect.origin.y < maxPoint.y;
             subRect.origin.y += 16) {
            for (subRect.origin.x = _location.origin.x; subRect.origin.x < maxPoint.x;
                 subRect.origin.x += 16) {
                subRect.size.width = 16;
                subRect.size.height = 16;

                if (maxPoint.y - subRect.origin.y < 16)
                    subRect.size.height = maxPoint.y - subRect.origin.y;
                if (maxPoint.x - subRect.origin.x < 16)
                    subRect.size.width = maxPoint.x - subRect.origin.x;

                [_subrectangles addObject:[[[RFBHextileSubrectangle alloc] initWithLocation:subRect superrectangle:self adaptor:adaptor] autorelease]];
            }
        }
    }

    return self;
}

- (void)dealloc
{
    [_subrectangles release];

    [super dealloc];
}

- (void)setLastBackgroundColor:(unsigned)color
{
    _lastBackgroundColor = color;
}

- (unsigned)lastBackgroundColor
{
    return _lastBackgroundColor;
}

- (void)drawWithFramebuffer:(VNCFramebuffer *)framebuffer
{
    NSEnumerator *rectEnumerator = [_subrectangles objectEnumerator];
    RFBHextileSubrectangle *subrect;

    while (subrect = [rectEnumerator nextObject])
        [subrect drawWithFramebuffer:framebuffer];
}

- (id)initWithCoder:(NSCoder *)coder
{
    if (self = [super initWithCoder:coder]) {
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_lastBackgroundColor];
        _subrectangles = [[coder decodeObject] retain];
    }

    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [super encodeWithCoder:coder];

    [coder encodeValueOfObjCType:@encode(unsigned) at:&_lastBackgroundColor];
    [coder encodeObject:_subrectangles];
}

@end

@implementation RFBHextileSubrectangle

- (id)initWithLocation:(NSRect)location superrectangle:(RFBHextileRectangle *)superrect adaptor:(RFBProtocolAdaptor *)adaptor
{
    if (self = [super init]) {
        _location = location;
        _subEncodingMask = [[adaptor socket] readByte];

        if (_subEncodingMask & RFBHextileRaw) {
            unsigned i = location.size.width*location.size.height*([adaptor bitsPerPixel]/8);
            _bitsPerPixel = [adaptor bitsPerPixel];
            _pixelData = [[[adaptor socket] readDataOfLength:i] retain];
        }
        else {
            if (_subEncodingMask & RFBHextileBackgroundSpecified) {
                _backgroundColor = [adaptor read24BitColor];
                [superrect setLastBackgroundColor:_backgroundColor];
            }
            else
                _backgroundColor = [superrect lastBackgroundColor];

            if (_subEncodingMask & RFBHextileForegroundSpecified)
                _foregroundColor = [adaptor read24BitColor];

            if (_subEncodingMask & RFBHextileAnySubrects) {
                NSRect *rectList = NULL;
                unsigned *colorList = NULL;
                octet i;

                _numberOfRects = [[adaptor socket] readByte];                
                _rectListCover = [[NSMutableData alloc] initWithLength:_numberOfRects*sizeof(NSRect)];
                rectList = [_rectListCover mutableBytes];

                if (_subEncodingMask & RFBHextileSubrectsColoured) {
                    _colorListCover = [[NSMutableData alloc] initWithLength:_numberOfRects*sizeof(unsigned)];
                    colorList = [_colorListCover mutableBytes];
                }
                
                for (i = 0; i < _numberOfRects; ++i) {
                    octet xy, hw;
                    if (_subEncodingMask & RFBHextileSubrectsColoured)
                        colorList[i] = [adaptor read24BitColor];

                    xy = [[adaptor socket] readByte];
                    hw = [[adaptor socket] readByte];

                    rectList[i].origin.x = _location.origin.x + (xy >> 4);
                    rectList[i].origin.y = _location.origin.y + (xy & 0xf);
                    rectList[i].size.width = ((hw >> 4)+1);
                    rectList[i].size.height = ((hw & 0xF)+1);
                }
            }
        }
    }

    return self;
}

- (void)dealloc
{
    [_pixelData release];
    [_rectListCover release];
    [_colorListCover release];

    [super dealloc];
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] location=%@ subEncodingMask=0x%02x>",
        [[self class] description], self,
        NSStringFromRect(_location), (int)_subEncodingMask];
}

- (void)drawWithFramebuffer:(VNCFramebuffer *)framebuffer
{
    if (_subEncodingMask & RFBHextileRaw) {
        [framebuffer compositePixelData:_pixelData
                             targetRect:_location
                           bitsPerPixel:_bitsPerPixel];
    }
    else {
        [framebuffer fillRect:_location color:_backgroundColor];

        if (_subEncodingMask & RFBHextileAnySubrects) {
            const NSRect *rectList = [_rectListCover bytes];
            const unsigned *colorList = [_colorListCover bytes];
            octet i;

            if (_subEncodingMask & RFBHextileForegroundSpecified)
                [framebuffer fillRectList:rectList colorList:colorList count:_numberOfRects];
            else
                for (i = 0; i < _numberOfRects; ++i)
                    [framebuffer fillRect:rectList[i] color:_foregroundColor];
        }
    }
}

- (id)initWithCoder:(NSCoder *)coder
{
    if (self = [super init]) {
        _location = [coder decodeRect];
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_subEncodingMask];
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_bitsPerPixel];
        _pixelData = [[coder decodeDataObject] retain];
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_backgroundColor];
        [coder decodeValueOfObjCType:@encode(unsigned) at:&_foregroundColor];
        [coder decodeValueOfObjCType:@encode(octet) at:&_numberOfRects];
        _rectListCover = [[coder decodeDataObject] retain];
        _colorListCover = [[coder decodeDataObject] retain];
    }

    return self;
}

- (void)encodeWithCoder:(NSCoder *)coder
{
    [coder encodeRect:_location];
    [coder encodeValueOfObjCType:@encode(unsigned) at:&_subEncodingMask];
    [coder encodeValueOfObjCType:@encode(unsigned) at:&_bitsPerPixel];
    [coder encodeDataObject:_pixelData];
    [coder encodeValueOfObjCType:@encode(unsigned) at:&_backgroundColor];
    [coder encodeValueOfObjCType:@encode(unsigned) at:&_foregroundColor];
    [coder encodeValueOfObjCType:@encode(octet) at:&_numberOfRects];
    [coder encodeDataObject:_rectListCover];
    [coder encodeDataObject:_colorListCover];
}


@end
