/*
 * VNCFramebuffer.m
 * A simple off-screen framebuffer optimized for drawing rectangles and raw images.
 *
 * Copyright (c) 1997 21st Century Software, New York City.
 *
 * See the file Reference/License.rtf for information on usage and
 * redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by David Young <dwy@ace.net>, 1997.
 */

#import "VNC.h"

@implementation VNCFramebuffer

- (id)initWithSize:(NSSize)size
{
    if (self = [super init]) {
        _width = size.width;
        _height = size.height;
        _bytesPerRow = size.width * sizeof(unsigned);
        _minimumUpdateRect = NSMakeRect(0, 0, 0, 0);
        _framebufferCover = [[NSMutableData alloc] initWithLength:_bytesPerRow*_width];
        _framebufferPtr = [_framebufferCover mutableBytes];
    }

    return self;
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<%@[0x%lx] %dx%d bytesPerRow=%d minUpdateRect=%@ framebufferPtr=0x%lx>",
        [[self class] description], self,
        _width, _height, _bytesPerRow, NSStringFromRect(_minimumUpdateRect),
        _framebufferPtr];
}

- (void)dealloc
{
    [_framebufferCover release];

    [super dealloc];
}

- (NSImage *)framebufferImage
{
    NSImage *image = [[[NSImage alloc] initWithSize:NSMakeSize(_width, _height)] autorelease];
    const unsigned char *pixelPointers[5];
    NSBitmapImageRep *imageRep;
    
    pixelPointers[0] = (const unsigned char *)_framebufferPtr;
    imageRep = [[[NSBitmapImageRep alloc]
        initWithBitmapDataPlanes:pixelPointers
                      pixelsWide:_width
                      pixelsHigh:_height
                   bitsPerSample:8
                 samplesPerPixel:3
                        hasAlpha:NO
                        isPlanar:NO
                  colorSpaceName:NSCalibratedRGBColorSpace
                     bytesPerRow:_bytesPerRow
                    bitsPerPixel:32] autorelease];

    [image addRepresentation:imageRep];

    return image;
}

- (NSData *)subPixelsInRect:(NSRect)rect
{
    unsigned bytesPerRow = rect.size.width * sizeof(unsigned);
    NSMutableData *subData = [NSMutableData dataWithLength:bytesPerRow*rect.size.height];
    unsigned *subDataPtr = [subData mutableBytes];
    unsigned x, y;

    for (y = rect.origin.y; y < rect.origin.y+rect.size.height; ++y)
        for (x = rect.origin.x; x < rect.origin.x+rect.size.width; ++x)
            *subDataPtr++ = _framebufferPtr[(y*_width)+x];

    return subData;
}

- (void)compositePixelData:(NSData *)pixelData targetRect:(NSRect)target bitsPerPixel:(unsigned)bitsPerPixel
{
    switch(bitsPerPixel) {
        case 32: {
            unsigned x, y;
            const unsigned *pixelPtr = [pixelData bytes];                          
            for (y = target.origin.y; y < (target.origin.y + target.size.height); ++y)
                for (x = target.origin.x; x < (target.origin.x + target.size.width); ++x)
                    _framebufferPtr[(y * _width) + x] = NSSwapBigIntToHost(*pixelPtr++);
            
            break;
        }

        case 16:
        case 8:
           [NSException raise:NSInvalidArgumentException
                       format:@"%@ unsupported bit depth %d",
               NSStringFromSelector(_cmd), bitsPerPixel];
            
        default:
            [NSException raise:NSInvalidArgumentException
                        format:@"%@ invalid bit depth %d",
                NSStringFromSelector(_cmd), bitsPerPixel];
    }

    _minimumUpdateRect = NSUnionRect(_minimumUpdateRect, target);
}

- (void)copyPixelsFromRect:(NSRect)srcRect toRect:(NSRect)destRect
{
    unsigned srcX, srcY;
    unsigned dstX = destRect.origin.x, dstY = destRect.origin.y;

    for (srcY = srcRect.origin.y; srcY < srcRect.origin.y+srcRect.size.height; ++srcY, ++dstY) {
        for (srcX = srcRect.origin.x; srcX < srcRect.origin.x+srcRect.size.width; ++srcX)
            _framebufferPtr[(dstY*_width) + dstX++] = _framebufferPtr[(srcY*_width) + srcX];

        dstX = destRect.origin.x;
    }

    _minimumUpdateRect = NSUnionRect(_minimumUpdateRect, destRect);
}

- (void)fillRect:(NSRect)rect color:(unsigned)color
{
    unsigned x, y;

    for (y = rect.origin.y; y < (rect.origin.y + rect.size.height); ++y)
        for (x = rect.origin.x; x < (rect.origin.x + rect.size.width); ++x)
            _framebufferPtr[(y * _width) + x] = color;
            
    _minimumUpdateRect = NSUnionRect(_minimumUpdateRect, rect);
}

- (void)fillRectList:(const NSRect *)rectList colorList:(const unsigned *)colorList count:(unsigned)count
{
    unsigned i = 0;

    while (i < count) {
        NSRect rect = rectList[i];
        unsigned x, y;
        unsigned color = colorList[i];

        for (y = rect.origin.y; y < (rect.origin.y + rect.size.height); ++y)
            for (x = rect.origin.x; x < (rect.origin.x + rect.size.width); ++x)
                _framebufferPtr[(y * _width) + x] = color;

        _minimumUpdateRect = NSUnionRect(_minimumUpdateRect, rect);
        ++i;
    }
}

- (void)updateRect:(NSRect)updateRect
{
    const unsigned char *pixelPointers[5];
    NSBitmapImageRep *imageRep;

    if (updateRect.size.width == _width && updateRect.size.height == _height)
        pixelPointers[0] = (const unsigned char *)_framebufferPtr;
    else
        pixelPointers[0] = [[self subPixelsInRect:updateRect] bytes];

    imageRep = [[[NSBitmapImageRep alloc]
        initWithBitmapDataPlanes:pixelPointers
                      pixelsWide:updateRect.size.width
                      pixelsHigh:updateRect.size.height
                   bitsPerSample:8
                 samplesPerPixel:3
                        hasAlpha:NO
                        isPlanar:NO
                  colorSpaceName:NSCalibratedRGBColorSpace
                     bytesPerRow:updateRect.size.width*sizeof(unsigned)
                    bitsPerPixel:32] autorelease];

    // correct for coordinate weirdness
    updateRect.origin.y = (_height/2)+((_height/2)-updateRect.origin.y);
    updateRect.origin.y -= updateRect.size.height;
    
    [imageRep drawAtPoint:updateRect.origin];
}

- (void)update
{
    [self updateRect:NSMakeRect(0, 0, _width, _height)];
    _minimumUpdateRect = NSMakeRect(0, 0, 0, 0);
}

- (void)updateMinimum
{
    [self updateRect:_minimumUpdateRect];
    _minimumUpdateRect = NSMakeRect(0, 0, 0, 0);
}

@end

